/***********************************************************************
Based on gpsdrive 
Copyright (c) 2001-2004 Fritz Ganter <ganter@ganter.at>

Website: www.gpsdrive.de

Disclaimer: Please do not use for navigation. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*********************************************************************/

#include<gtk/gtk.h>
#include<iostream>
#include"MyColors.h"
#include"Map.h"
#include"ValueField.h"

class MainWindow : public RedrawEventListener
{
public:
    MainWindow();
    /* Copy Image from loaded map */
    static gboolean expose_cb (GtkWidget *widget,GdkEventExpose *event);
    static gint mapclick_cb(GtkWidget *widget, GdkEventButton *event);
    static gint quit_program(GtkWidget *widget, gpointer datum);

    gint error_popup (gpointer datum);
    void loadmap(char *path);
    static MainWindow *instance();
    GdkDrawable *getPixmap() { return drawable_; }

    // interface Redrawlistener
    virtual void redraw();

 private:
    void initWindow();
    void initGraphicKontext();
    static void allocColors(GtkWidget *mainwindow);
    void fillwithColor();
    bool calcdrawable();
    void calcinitialscale(Map *m);
    /* groesse des Drawing-Areas */
    gint real_screen_x_;
    gint real_screen_y_;


    /*---- Widgets ---------*/
    GtkWidget *mainwindow_;
    GtkWidget *quit_;
    GtkWidget *dframe; // Area for map
    GtkWidget *map_drawing_area_; 
    DoubleValueField scalevf_;
    IntValueField xoffvf_;
    IntValueField yoffvf_;
    GdkGC *kontext_;    
    GdkDrawable *drawable_; // bild 1:1 zum zeichnen

    //* state *
    Map *themap_;  // geladene Landkarte
    bool exposed;
    GdkPixbuf *tempimage_; // zu calcdrawable  Renderer ?
    
    bool iszoomed;

    MyColors *mycolors_;
    static MainWindow *instance_;
};
MainWindow *MainWindow::instance_=0;

MainWindow *MainWindow::instance()
{
    if(instance_==0)
    {
	instance_=new MainWindow();
    }
    return instance_;
}

MainWindow::MainWindow():real_screen_x_(640),real_screen_y_(512),
			 themap_(NULL),exposed(false),
			 tempimage_(NULL)
{
    initWindow();
    tempimage_ = gdk_pixbuf_new (GDK_COLORSPACE_RGB, 0, 8, 1280, 1024);
}

void 
MainWindow::initWindow()
{
#define PADDING 1 // int_padding ???

    GtkWidget *hbig;
    GtkWidget *vmenulist;
    mainwindow_ = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    const char *maintitle = "Simple map display";
    gtk_window_set_title(GTK_WINDOW(mainwindow_), maintitle);
    gtk_container_set_border_width(GTK_CONTAINER(mainwindow_), 0);
    gtk_window_set_position(GTK_WINDOW(mainwindow_), GTK_WIN_POS_CENTER);
    gtk_signal_connect(GTK_OBJECT(mainwindow_), "delete_event",
			GTK_SIGNAL_FUNC(quit_program), NULL);


    hbig = gtk_hbox_new(FALSE, 0 * PADDING);

    /*  Area for map */
    dframe = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(dframe), GTK_SHADOW_IN);
    map_drawing_area_ = gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(map_drawing_area_), real_screen_x_, real_screen_y_);
    
    gtk_container_add(GTK_CONTAINER(dframe), map_drawing_area_);
    
    gtk_box_pack_start(GTK_BOX(hbig), dframe, TRUE, TRUE, 0 * PADDING);
    gtk_widget_add_events(GTK_WIDGET(map_drawing_area_), GDK_BUTTON_PRESS_MASK);
    gtk_signal_connect (GTK_OBJECT (map_drawing_area_),
    	      "expose_event", GTK_SIGNAL_FUNC (expose_cb),NULL);

    gtk_signal_connect_object(GTK_OBJECT(map_drawing_area_),
			       "button-press-event",
			       GTK_SIGNAL_FUNC(mapclick_cb),
			       GTK_OBJECT(map_drawing_area_));

    
    /* ------ Horizontal Menulist on the right side ----------*/
    vmenulist = gtk_vbox_new(FALSE, 0);

    /*-scale-*/
    scalevf_.init(GTK_BOX(vmenulist),"Scale",1.0);
    scalevf_.addRedrawEventListener(this);

    /*-xoff und yoff-*/
    xoffvf_.init(GTK_BOX(vmenulist),"Xoff in pix",0);
    xoffvf_.addRedrawEventListener(this);
    yoffvf_.init(GTK_BOX(vmenulist),"Yoff in pix",0);
    yoffvf_.addRedrawEventListener(this);

    /*  Quit button */
    quit_ = gtk_button_new_from_stock(GTK_STOCK_QUIT);
    gtk_button_set_use_underline(GTK_BUTTON(quit_), TRUE);
    
    gtk_signal_connect(GTK_OBJECT(quit_),
			"clicked", GTK_SIGNAL_FUNC(quit_program), 0);
    
    gtk_container_add(GTK_CONTAINER(mainwindow_),hbig);
    gtk_box_pack_start(GTK_BOX(vmenulist), quit_, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbig),vmenulist,TRUE, TRUE, 0);
    gtk_widget_show_all(mainwindow_);

    mycolors_ = new MyColors(mainwindow_);
    /*------ Initialisierung graphikkontext und drawable ----*/
    drawable_ = gdk_pixmap_new(mainwindow_->window, real_screen_x_, real_screen_y_, -1);
    initGraphicKontext();
    fillwithColor();
}


void MainWindow::initGraphicKontext()
{
    GdkRectangle rectangle =
    {
	0, 0,  real_screen_x_, real_screen_y_
    };
    /*------ Initialisierung graphikkontext ----*/
    kontext_ = gdk_gc_new(mainwindow_->window);
    gdk_gc_set_clip_origin(kontext_, 0, 0);
    gdk_gc_set_clip_rectangle(kontext_, &rectangle);
    gdk_gc_set_function(kontext_, GDK_COPY);
}

void 
MainWindow::fillwithColor()
{
    GdkRectangle rectangle =
    {
	0, 0,  real_screen_x_, real_screen_y_
    };
    /*------ Initialisierung graphikkontext und drawable ----*/
    GdkGC *kontext = gdk_gc_new(mainwindow_->window);
    gdk_gc_set_clip_origin(kontext, 0, 0);
    gdk_gc_set_clip_rectangle(kontext, &rectangle);
    gdk_gc_set_function(kontext, GDK_COPY);
    gdk_gc_set_foreground(kontext, mycolors_->lcd2());
    gdk_draw_rectangle (drawable_, kontext, TRUE, 0, 0,
			real_screen_x_,real_screen_y_);
}

gboolean
MainWindow::expose_cb(GtkWidget *widget, GdkEventExpose *event)
{
    gdk_draw_drawable(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    MainWindow::instance()->getPixmap(),
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);
    return FALSE;
}

void
MainWindow::calcinitialscale(Map *m)
{
    double scale_x = (double)real_screen_x_/(double)(m->getWidth());
    double scale_y = (double)real_screen_y_/(double)(m->getHeight());
    double scale=(scale_x < scale_y ? scale_x : scale_y);
    std::cout << "Scale (" << scale_x << "," << scale_y << ") -> scale="
	      << scale << std::endl;
    scalevf_.set(scale);
}

void
MainWindow::redraw()
{
    std::cerr << "redraw ausgeloesst " << std::endl;
    (void)calcdrawable();
    //??? Question: can I tell the whole window to do a refresh ?
    gdk_draw_drawable(map_drawing_area_->window,
		      map_drawing_area_->style->fg_gc[GTK_WIDGET_STATE (map_drawing_area_)],
		      MainWindow::instance()->getPixmap(),
		      0, 0,
		      0, 0,
		      real_screen_x_,real_screen_y_);
}

bool
MainWindow::calcdrawable()
{
    std::cerr << "calcdrawable" << std::endl;
    //if (exposed) return TRUE;
    if(themap_)
    {
	int offx=-xoffvf_.get();
	int offy=-yoffvf_.get();
	//int offx=real_screen_x_-xoff-(int)(real_screen_x_*scalevf_.get());
	//int offy=real_screen_y_-yoff-(int)(real_screen_y_*scalevf_.get());
	gdk_pixbuf_scale(themap_->getImage(), tempimage_,
			 0, 0, /* dest x,y*/
			 real_screen_x_,real_screen_y_, /*dest w,h*/
			 offx,offy,  /* offset(x,y)*/
			 scalevf_.get(), scalevf_.get(), /* scale x,y */
			 GDK_INTERP_BILINEAR);
	gdk_draw_pixbuf(drawable_, kontext_, tempimage_,
			0, 0,
			0, 0,
			real_screen_x_, real_screen_y_,
			GDK_RGB_DITHER_NONE, 0, 0);
    }
    else
    {
	fillwithColor();
	std::cerr << "No Map " << std::endl;
    }
    exposed = TRUE;
    return TRUE;
}

gint
MainWindow::error_popup(gpointer datum)
{
  GtkWidget *popup;
  GtkWidget *knopf2;
  GtkWidget *knopf3;
  GtkWidget *image;

  popup = gtk_dialog_new();
  gtk_window_set_transient_for(GTK_WINDOW(popup), GTK_WINDOW(mainwindow_));

  image =
    gtk_image_new_from_stock(GTK_STOCK_DIALOG_ERROR, GTK_ICON_SIZE_DIALOG);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), image,
		      FALSE, FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(popup), 5);

  gtk_window_set_title(GTK_WINDOW(popup), "  Message  ");
  knopf2 = gtk_button_new_from_stock(GTK_STOCK_OK);
  GTK_WIDGET_SET_FLAGS(knopf2, GTK_HAS_FOCUS);

  gtk_signal_connect_object(GTK_OBJECT(knopf2), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(popup));

  gdk_beep();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->action_area), knopf2, TRUE,
		      TRUE, 20);
  knopf3 = gtk_label_new((const char *)datum);
  gtk_label_set_use_markup(GTK_LABEL(knopf3), TRUE);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(popup)->vbox), knopf3, FALSE,
		      FALSE, 10);


  gtk_window_set_position(GTK_WINDOW(popup), GTK_WIN_POS_CENTER);
  gtk_widget_show(image);
  gtk_widget_show_all(popup);

  return(FALSE);
}

void MainWindow::loadmap(char *path)
{
    if(themap_)
    {
	delete themap_;
	themap_=0;
    }

    try
    {
	themap_=new Map(path);
    }
    catch(std::exception &e)
    {
	error_popup((gpointer *) e.what());
	themap_=0;
    }

    iszoomed = FALSE;
    xoffvf_.set(0);
    yoffvf_.set(0);
    if(themap_)
    {
	calcinitialscale(themap_);
    }
    calcdrawable();
}

gint
MainWindow::quit_program(GtkWidget *widget, gpointer datum)
{
  std::cerr << "Und Tschuess" << std::endl;
  gtk_main_quit();
  return(TRUE);
}

gint
MainWindow::mapclick_cb(GtkWidget *widget, GdkEventButton *event)
{
    gdouble x, y;
    // double lon, lat, vali;
    GdkModifierType state;
    if (event->button)
    {
	gint xi,yi;
	gdk_window_get_pointer (event->window, &xi, &yi, &state);
	x=(gdouble)xi;
	y=(gdouble)yi;
    }
    else
    {
      x = event->x;
      y = event->y;
      //state = event->state;
    }
    std::cout << "mapclick_cb: x=" << x << " y=" << y 
	      << " (state=" << state << ")" << std::endl;
    return TRUE;
}

int 
main(int argc,char **argv)
{
    /*    Setting locale for correct Umlauts */
    gtk_set_locale();
    /*  initialization for GTK+ */
    gtk_init(&argc, &argv);
    /* Needed 4 hours to find out that this is IMPORTANT!!!! */
    gdk_rgb_init();

    MainWindow::instance();
    char **arg=argv+1;
    char *mapfile=0;
    while(*arg!=0)
    {
	if(strcmp(*arg,"-map")==0)
	{
	    mapfile=*++arg;
	}
	arg++;
    }
    if(mapfile!=0)
    {
	std::cerr << "Loading Mapfile " << mapfile << std::endl;
	MainWindow::instance()->loadmap(mapfile);
    }
    
    /*  Mainloop */
    gtk_main();
}
